@@ -1,4 +1,3 @@
-/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /* GIO - GLib Input, Output and Streaming Library
  * 
  * Copyright (C) 2008 Benjamin Otte <otte@gnome.org>
@@ -135,6 +134,8 @@ G_DEFINE_TYPE (GVfsBackendFtp, g_vfs_backend_ftp, G_VFS_TYPE_BACKEND)
 
 #define STATUS_GROUP(status) ((status) / 100)
 
+typedef void (* Ftp550Handler) (FtpConnection *conn, const FtpFile *file);
+
 /*** FTP CONNECTION ***/
 
 struct _FtpConnection
@@ -250,8 +251,8 @@ ftp_connection_set_error_from_response (FtpConnection *conn, guint response)
       case 550: /* Requested action not taken. File unavailable (e.g., file not found, no access). */
 	/* FIXME: This is a lot of different errors. So we have to pretend to 
 	 * be smart here. */
-	code = G_IO_ERROR_NOT_FOUND;
-	msg = _("File unavailable");
+	code = G_IO_ERROR_FAILED;
+	msg = _("Operation failed");
 	break;
       case 451: /* Requested action aborted: local error in processing. */
 	code = G_IO_ERROR_FAILED;
@@ -289,7 +290,7 @@ ftp_connection_set_error_from_response (FtpConnection *conn, guint response)
     }
 
   DEBUG ("error: %s\n", msg);
-  g_set_error (&conn->error, G_IO_ERROR, code, msg);
+  g_set_error_literal (&conn->error, G_IO_ERROR, code, msg);
 }
 
 /**
@@ -298,6 +299,7 @@ ftp_connection_set_error_from_response (FtpConnection *conn, guint response)
  * RESPONSE_PASS_300: Don't treat 3XX responses, but return them
  * RESPONSE_PASS_400: Don't treat 4XX responses, but return them
  * RESPONSE_PASS_500: Don't treat 5XX responses, but return them
+ * RESPONSE_PASS_550: Don't treat 550 responses, but return them
  * RESPONSE_FAIL_200: Fail on a 2XX response
  */
 
@@ -306,7 +308,8 @@ typedef enum {
   RESPONSE_PASS_300 = (1 << 1),
   RESPONSE_PASS_400 = (1 << 2),
   RESPONSE_PASS_500 = (1 << 3),
-  RESPONSE_FAIL_200 = (1 << 4)
+  RESPONSE_PASS_550 = (1 << 4),
+  RESPONSE_FAIL_200 = (1 << 5)
 } ResponseFlags;
 
 /**
@@ -357,8 +360,8 @@ ftp_connection_receive (FtpConnection *conn,
 	    }
 	  else
 	    {
-	      g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-			   _("Invalid reply"));
+	      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+				   _("Invalid reply"));
 	      return 0;
 	    }
 	}
@@ -386,8 +389,8 @@ ftp_connection_receive (FtpConnection *conn,
 	      break;
 	    if (n_bytes > 0)
 	      continue;
-	    g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-			 _("Invalid reply"));
+	    g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+				 _("Invalid reply"));
 	    /* fall through */
 	  case SOUP_SOCKET_ERROR:
 	    conn->read_buffer[conn->read_bytes] = 0;
@@ -405,8 +408,8 @@ ftp_connection_receive (FtpConnection *conn,
 	      last_line[1] < '0' || last_line[1] > '9' ||
 	      last_line[2] < '0' || last_line[2] > '9')
 	    {
-	      g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-			   _("Invalid reply"));
+	      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+				   _("Invalid reply"));
 	      return 0;
 	    }
 	  response = 100 * (last_line[0] - '0') +
@@ -418,8 +421,8 @@ ftp_connection_receive (FtpConnection *conn,
 	    reply_state = MULTILINE;
 	  else
 	    {
-	      g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-			   _("Invalid reply"));
+	      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+				   _("Invalid reply"));
 	      return 0;
 	    }
 	}
@@ -460,7 +463,7 @@ ftp_connection_receive (FtpConnection *conn,
 	return 0;
 	break;
       case 5:
-	if (flags & RESPONSE_PASS_500)
+	if ((flags & RESPONSE_PASS_500) || (response == 550 && (flags & RESPONSE_PASS_550)))
 	  break;
 	ftp_connection_set_error_from_response (conn, response);
 	return 0;
@@ -525,8 +528,8 @@ ftp_connection_sendv (FtpConnection *conn,
       case SOUP_SOCKET_EOF:
 	if (n_bytes == command->len)
 	  break;
-	g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-	    _("broken transmission"));
+	g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+			     _("broken transmission"));
 	/* fall through */
       case SOUP_SOCKET_ERROR:
 	g_string_free (command, TRUE);
@@ -564,6 +567,57 @@ ftp_connection_send (FtpConnection *conn,
   return response;
 }
 
+static void
+ftp_connection_check_file (FtpConnection *conn,
+                           const Ftp550Handler *handlers,
+                           const FtpFile *file)
+{
+  while (*handlers && !ftp_connection_in_error (conn))
+    {
+      (*handlers) (conn, file);
+      handlers++;
+    }
+}
+
+static guint
+ftp_connection_send_and_check (FtpConnection *conn,
+                               ResponseFlags flags,
+                               const Ftp550Handler *handlers,
+                               const FtpFile *file,
+                               const char *format,
+                               ...) G_GNUC_PRINTF (5, 6);
+static guint
+ftp_connection_send_and_check (FtpConnection *conn,
+                               ResponseFlags flags,
+                               const Ftp550Handler *handlers,
+                               const FtpFile *file,
+                               const char *format,
+                               ...)
+{
+  va_list varargs;
+  guint response;
+
+  /* check that there's no 550 handling used - don't allow bad use of API */
+  g_return_val_if_fail ((flags & RESPONSE_PASS_550) == 0, 0);
+  g_return_val_if_fail (handlers != NULL, 0);
+  g_return_val_if_fail (file != NULL, 0);
+
+  va_start (varargs, format);
+  response = ftp_connection_sendv (conn,
+                                   flags | RESPONSE_PASS_550,
+                                   format,
+                                   varargs);
+  va_end (varargs);
+  if (response == 550)
+    {
+      ftp_connection_check_file (conn, handlers, file);
+      if (!ftp_connection_in_error (conn))
+          ftp_connection_set_error_from_response (conn, response);
+      response = 0;
+    }
+  return response;
+}
+
 static void
 ftp_connection_parse_features (FtpConnection *conn)
 {
@@ -617,10 +671,10 @@ ftp_connection_create (SoupAddress * addr,
   if (!SOUP_STATUS_IS_SUCCESSFUL (status))
     {
       /* FIXME: better error messages depending on status please */
-      g_set_error (&conn->error,
-		   G_IO_ERROR,
-		   G_IO_ERROR_HOST_NOT_FOUND,
-		   _("Could not connect to host"));
+      g_set_error_literal (&conn->error,
+			   G_IO_ERROR,
+			   G_IO_ERROR_HOST_NOT_FOUND,
+			   _("Could not connect to host"));
     }
 
   ftp_connection_receive (conn, 0);
@@ -759,8 +813,8 @@ ftp_connection_ensure_data_connection (FtpConnection *conn)
     }
   if (*s == 0)
     {
-      g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
-		   _("Invalid reply"));
+      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FAILED,
+			   _("Invalid reply"));
       return FALSE;
     }
   ip = g_strdup_printf ("%u.%u.%u.%u", ip1, ip2, ip3, ip4);
@@ -777,10 +831,10 @@ have_address:
   if (!SOUP_STATUS_IS_SUCCESSFUL (status))
     {
       /* FIXME: better error messages depending on status please */
-      g_set_error (&conn->error,
-		   G_IO_ERROR,
-		   G_IO_ERROR_HOST_NOT_FOUND,
-		   _("Could not connect to host"));
+      g_set_error_literal (&conn->error,
+			   G_IO_ERROR,
+			   G_IO_ERROR_HOST_NOT_FOUND,
+			   _("Could not connect to host"));
       g_object_unref (conn->data);
       conn->data = NULL;
       return FALSE;
@@ -853,9 +907,9 @@ ftp_connection_cd (FtpConnection *conn, const FtpFile *file)
 					"CWD %s", file);
   if (response == 550)
     {
-      g_set_error (&conn->error, 
-	           G_IO_ERROR, G_IO_ERROR_NOT_DIRECTORY,
-		   _("The file is not a directory"));
+      g_set_error_literal (&conn->error, 
+		           G_IO_ERROR, G_IO_ERROR_NOT_DIRECTORY,
+			   _("The file is not a directory"));
       response = 0;
     }
   else if (STATUS_GROUP (response) == 5)
@@ -1304,8 +1358,8 @@ do_mount (GVfsBackend *backend,
 		        &password_save) ||
 	  aborted) 
 	{
-	  g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
-		       "%s", _("Password dialog cancelled"));
+	  g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
+			       _("Password dialog cancelled"));
 	  break;
 	}
 
@@ -1461,6 +1515,21 @@ do_unmount (GVfsBackend *   backend,
   g_vfs_job_succeeded (G_VFS_JOB (job));
 }
 
+static void
+error_550_is_directory (FtpConnection *conn, const FtpFile *file)
+{
+  guint response = ftp_connection_send (conn, 
+                                        RESPONSE_PASS_550,
+                                        "CWD %s", file);
+
+  if (STATUS_GROUP (response) == 2)
+    {
+      g_set_error_literal (&conn->error, G_IO_ERROR, 
+	                   G_IO_ERROR_IS_DIRECTORY,
+        	           _("File is directory"));
+    }
+}
+
 static void
 do_open_for_read (GVfsBackend *backend,
 		  GVfsJobOpenForRead *job,
@@ -1469,6 +1538,7 @@ do_open_for_read (GVfsBackend *backend,
   GVfsBackendFtp *ftp = G_VFS_BACKEND_FTP (backend);
   FtpConnection *conn;
   FtpFile *file;
+  static const Ftp550Handler open_read_handlers[] = { error_550_is_directory, NULL };
 
   conn = g_vfs_backend_ftp_pop_connection (ftp, G_VFS_JOB (job));
   if (!conn)
@@ -1477,9 +1547,11 @@ do_open_for_read (GVfsBackend *backend,
   ftp_connection_ensure_data_connection (conn);
 
   file = ftp_filename_from_gvfs_path (conn, filename);
-  ftp_connection_send (conn,
-		       RESPONSE_PASS_100 | RESPONSE_FAIL_200,
-		       "RETR %s", file);
+  ftp_connection_send_and_check (conn,
+		                 RESPONSE_PASS_100 | RESPONSE_FAIL_200, 
+		                 &open_read_handlers[0],
+		                 file,
+		                 "RETR %s", file);
   g_free (file);
 
   if (ftp_connection_in_error (conn))
@@ -1621,10 +1693,10 @@ do_create (GVfsBackend *backend,
   if (info)
     {
       g_object_unref (info);
-      g_set_error (&conn->error,
-	           G_IO_ERROR,
-		   G_IO_ERROR_EXISTS,
-		   _("Target file already exists"));
+      g_set_error_literal (&conn->error,
+		           G_IO_ERROR,
+			   G_IO_ERROR_EXISTS,
+			   _("Target file already exists"));
       goto error;
     }
   file = ftp_filename_from_gvfs_path (conn, filename);
@@ -1752,8 +1824,8 @@ do_enumerate_directory (FtpConnection *conn)
 	{
 	  if (size >= 16384)
 	    {
-	      g_set_error (&conn->error, G_IO_ERROR, G_IO_ERROR_FILENAME_TOO_LONG,
-		           _("filename too long"));
+	      g_set_error_literal (&conn->error, G_IO_ERROR, G_IO_ERROR_FILENAME_TOO_LONG,
+			           _("filename too long"));
 	      break;
 	    }
 	  size += 128;
@@ -2023,10 +2095,10 @@ do_query_info (GVfsBackend *backend,
       g_object_unref (real);
     }
   else if (!ftp_connection_in_error (conn))
-    g_set_error (&conn->error,
-		 G_IO_ERROR,
-		 G_IO_ERROR_NOT_FOUND,
-		 _("File doesn't exist"));
+    g_set_error_literal (&conn->error,
+			 G_IO_ERROR,
+			 G_IO_ERROR_NOT_FOUND,
+			 _("File doesn't exist"));
 
   g_vfs_backend_ftp_push_connection (ftp, conn);
 }
@@ -2133,10 +2205,10 @@ do_set_display_name (GVfsBackend *backend,
   now = ftp_filename_construct (conn, dir, display_name);
   if (now == NULL)
     {
-      g_set_error (&conn->error, 
-	           G_IO_ERROR,
-	           G_IO_ERROR_INVALID_FILENAME,
-		   _("Invalid filename"));
+      g_set_error_literal (&conn->error, 
+		           G_IO_ERROR,
+		           G_IO_ERROR_INVALID_FILENAME,
+			   _("Invalid filename"));
     }
   ftp_connection_send (conn,
 		       RESPONSE_PASS_300 | RESPONSE_FAIL_200,
@@ -2186,10 +2258,10 @@ do_delete (GVfsBackend *backend,
 	  if (files)
 	    {
 	      g_static_rw_lock_reader_unlock (&ftp->directory_cache_lock);
-	      g_set_error (&conn->error, 
-			   G_IO_ERROR,
-			   G_IO_ERROR_NOT_EMPTY,
-			   "%s", g_strerror (ENOTEMPTY));
+	      g_set_error_literal (&conn->error, 
+				   G_IO_ERROR,
+				   G_IO_ERROR_NOT_EMPTY,
+				   g_strerror (ENOTEMPTY));
 	    }
 	  else
 	    ftp_connection_set_error_from_response (conn, response);
@@ -2268,9 +2340,9 @@ do_move (GVfsBackend *backend,
 
       g_free (basename);
       if (real == NULL)
-	g_set_error (&conn->error, 
-	             G_IO_ERROR, G_IO_ERROR_INVALID_FILENAME,
-		     _("Invalid destination filename"));
+	g_set_error_literal (&conn->error, 
+		             G_IO_ERROR, G_IO_ERROR_INVALID_FILENAME,
+			     _("Invalid destination filename"));
       else
 	{
 	  g_free (destfile);
@@ -2287,10 +2359,10 @@ do_move (GVfsBackend *backend,
       if (info)
 	{
 	  g_object_unref (info);
-	  g_set_error (&conn->error,
-		       G_IO_ERROR,
-	               G_IO_ERROR_EXISTS,
-		       _("Target file already exists"));
+	  g_set_error_literal (&conn->error,
+			       G_IO_ERROR,
+		               G_IO_ERROR_EXISTS,
+			       _("Target file already exists"));
 	  goto out;
 	}
     }
